open Extracted.PropQuantifiers
open Extracted.Formulas
open Extracted.Environments
open Extracted.Datatypes

(* for random testing. *)
open ExEnum
open Sys

let rec int_of_nat = function
| O -> 0
| S n -> 1 + int_of_nat n

let rec nat_of_int (n: int) =
    if n <= 0 then O
    else S (nat_of_int (n-1))

let rec string_of_formula = function
| Var v -> "x" ^ string_of_int (int_of_nat v)
| Bot -> "⊥"
| And (f, g) -> bracket f ^ " ∧ " ^ bracket g
| Or (f, g) -> bracket f ^ " ∨ " ^ bracket g
| Implies (Bot, _) -> "⊤"                             (* ⊥ -> φ ≡ ⊤ *)
| Implies (Implies(Bot, _), f) -> string_of_formula f (* ⊤ -> φ ≡ φ *)
| Implies (f, Bot) -> "¬" ^ bracket f       (* φ -> ⊥ ≡ ¬ φ *)
| Implies (_, Implies(Bot, _)) -> "⊤"                 (* φ -> ⊤ ≡ ⊤ *)
| Implies (f, g) -> if f = g then "⊤" else            (* φ -> φ ≡ ⊤ *)
                    bracket f ^ " → " ^ bracket g
and bracket e = match e with
| Var _ | Bot -> string_of_formula e
| e -> "(" ^ string_of_formula e ^ ")"


(* existential quantifier with respect to variable v*)
let e_var v f = coq_Ef (nat_of_int v) f;;
let e f = e_var 0 f

(* universal quantifier wrt v *)
let a_var v f = coq_Af (nat_of_int v) f;;
let a f = a_var 0 f

(* n first variables *)
let rec n_vars n = if n <= 1 then [O] else O :: List.map (fun x -> S x) (n_vars (n - 1))

let e_vars = from_list ~name:"variables" (n_vars 3)

(* Type term is recursive, hence we need a lazy enumeration first. *)
let rec l_e_formulas = lazy
    begin
      let r_formulas = pay l_e_formulas in
      union
	[ single Bot ;
      map e_vars (fun x -> Var x) ;
	  map (pair r_formulas r_formulas) (fun (t1, t2) -> And (t1, t2)) ;
      map (pair r_formulas r_formulas) (fun (t1, t2) -> Or (t1, t2)) ;
      map (pair r_formulas r_formulas) (fun (t1, t2) -> Implies (t1, t2)) ;
	]
    end


(* Enumeration for formulas. *)
let e_formulas = Lazy.force l_e_formulas

(* The code below prints _num_ formulas and their existential interpolant *)

let nb_args = Array.length Sys.argv

let start = if nb_args < 2 then 0 else int_of_string Sys.argv.(1)
let num = if nb_args < 3 then 1 else int_of_string Sys.argv.(2)

let usage_string =
"Usage: propquant s n shows n example formulas φ(0), ... φ(n-1),
        and computes E x0 φ(i) and A x0 φ(i) for each 0 ≤ i < n.
        The formulas are chosen from an enumeration of all formulas in variables x0, x1, x2, x3, x4.
        The argument s determines where to start in this enumeration.

        When s is negative, a custom benchmark is tested instead, defined by the recursion:
        φ(0) = x(0)
        φ(n+1) = φ(n) → x(n+1)

        For example, try: propquant.exe -3, propquant.exe 0 10, propquant.exe 5893 100.\n"

let var (n: int) = Var (nat_of_int n)
let rec test_formula (n: int) =
    if n = 0 then var 0 else
    Implies (test_formula(n-1), var(n))
(* let test_formula (n: int) =
  if n = 1 then Implies(Implies(var 1, var 0), var 0) else Bot *)

let show_test (v: int) (f: form) =
    print_string("φ        : " ^ string_of_formula f ^ "\n" ^
                 "Weight: " ^ string_of_int (int_of_nat (weight f)) ^ "\n");
    let t_start = Sys.time() in
    let e_result = e_var v f in
    let run_time = Sys.time() -. t_start in
    print_string("Time: " ^ string_of_float run_time ^ "s\n");
    print_string("E_(x" ^ string_of_int v ^ ") φ : " ^
                  string_of_formula e_result ^ "\n" ^
                 "Weight: " ^ string_of_int (int_of_nat (weight e_result)) ^ "\n");
    let a_result = a_var v f in
    print_string("A_(x" ^ string_of_int v ^ ") φ : " ^
                 string_of_formula a_result ^ "\n" ^
                "Weight: " ^ string_of_int (int_of_nat (weight a_result)) ^ "\n-------------------\n");;

let custom_test (n: int) =
  print_string ("Performing custom benchmark " ^ (string_of_int n) ^ " times\n----------------\n");
  for i = 1 to n do
    show_test 0 (test_formula i);
  done;;

let () = let do_it =
         if nb_args < 2 then (print_string usage_string)
         else
            if start < 0 then (custom_test (0-start))
            else
            (print_string ("Formula:\n");
            show e_formulas string_of_formula start num;
            print_string ("\nE_(x0) (Formula):\n");
            show (ExEnum.map e_formulas e) string_of_formula start num;
            print_string ("\nA_(x0) (Formula):\n");
            show (ExEnum.map e_formulas a) string_of_formula start num;
            print_string ("-------\n"))
         in
            do_it
